# coding: utf-8

# docs
from __future__ import annotations
from typing import List, NamedTuple, Optional, Dict

from supervisely.api.module_api import ApiField, ModuleApi, RemoveableBulkModuleApi
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely._utils import batched


class FigureApi(RemoveableBulkModuleApi):
    """
    Figure object for :class:`VideoAnnotation<supervisely.video_annotation.video_annotation.VideoAnnotation>`.
    """
    @staticmethod
    def info_sequence():
        """
        NamedTuple FigureInfo information about Figure.

        :Example:

         .. code-block:: python

            FigureInfo(id=588801373,
                       updated_at='2020-12-22T06:37:13.183Z',
                       created_at='2020-12-22T06:37:13.183Z',
                       entity_id=186648101,
                       object_id=112482,
                       project_id=110366,
                       dataset_id=419886,
                       frame_index=0,
                       geometry_type='bitmap',
                       geometry={'bitmap': {'data': 'eJwdlns8...Cgj4=', 'origin': [335, 205]}})
        """
        return [ApiField.ID,
                ApiField.UPDATED_AT,
                ApiField.CREATED_AT,
                ApiField.ENTITY_ID,
                ApiField.OBJECT_ID,
                ApiField.PROJECT_ID,
                ApiField.DATASET_ID,
                ApiField.FRAME_INDEX,
                ApiField.GEOMETRY_TYPE,
                ApiField.GEOMETRY
                ]

    @staticmethod
    def info_tuple_name():
        """"""
        return "FigureInfo"

    def get_info_by_id(self, id: int) -> NamedTuple:
        """
        Get Figure information by ID.

        :param id: Figure ID in Supervisely.
        :type id: int
        :return: Information about Figure. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`NamedTuple`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            figure_id = 588801373

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            figure_info = api.video.figure.get_info_by_id(figure_id)
            print(figure_info)
            # Output: [
            #     588801373,
            #     "2020-12-22T06:37:13.183Z",
            #     "2020-12-22T06:37:13.183Z",
            #     186648101,
            #     112482,
            #     110366,
            #     419886,
            #     0,
            #     "bitmap",
            #     {
            #         "bitmap": {
            #             "data": "eJw...Cgj4=",
            #             "origin": [
            #                 335,
            #                 205
            #             ]
            #         }
            #     }
            # ]
        """
        return self._get_info_by_id(id, 'figures.info')

    def create(self, entity_id: int, object_id: int, meta: Dict, geometry_json: Dict, geometry_type, track_id: int=None):
        """"""
        input_figure = {
            ApiField.META: meta,
            ApiField.OBJECT_ID: object_id,
            ApiField.GEOMETRY_TYPE: geometry_type,
            ApiField.GEOMETRY: geometry_json,
        }

        if track_id is not None:
            input_figure[ApiField.TRACK_ID] = track_id

        body = {ApiField.ENTITY_ID: entity_id, ApiField.FIGURES: [input_figure]}

        response = self._api.post("figures.bulk.add", body)
        return response.json()[0][ApiField.ID]

    def get_by_ids(self, dataset_id: int, ids: List[int]) -> List[NamedTuple]:
        """
        Get Figures information by IDs from given dataset ID.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param ids: List of Figures IDs.
        :type ids: List[int]
        :return: List of information about Figures. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[NamedTuple]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            dataset_id = 466642
            figures_ids = [642155547, 642155548, 642155549]
            figures_infos = api.video.figure.get_by_ids(dataset_id, figures_ids)
            print(figures_infos)
            # Output: [
            #     [
            #         642155547,
            #         "2021-03-23T13:25:34.705Z",
            #         "2021-03-23T13:25:34.705Z",
            #         198703211,
            #         152118,
            #         124976,
            #         466642,
            #         0,
            #         "rectangle",
            #         {
            #             "points": {
            #                 "exterior": [
            #                     [
            #                         2240,
            #                         1041
            #                     ],
            #                     [
            #                         2463,
            #                         1187
            #                     ]
            #                 ],
            #                 "interior": []
            #             }
            #         }
            #     ],
            #     [
            #         642155548,
            #         "2021-03-23T13:25:34.705Z",
            #         "2021-03-23T13:25:34.705Z",
            #         198703211,
            #         152118,
            #         124976,
            #         466642,
            #         1,
            #         "rectangle",
            #         {
            #             "points": {
            #                 "exterior": [
            #                     [
            #                         2248,
            #                         1048
            #                     ],
            #                     [
            #                         2455,
            #                         1176
            #                     ]
            #                 ],
            #                 "interior": []
            #             }
            #         }
            #     ],
            #     [
            #         642155549,
            #         "2021-03-23T13:25:34.705Z",
            #         "2021-03-23T13:25:34.705Z",
            #         198703211,
            #         152118,
            #         124976,
            #         466642,
            #         2,
            #         "rectangle",
            #         {
            #             "points": {
            #                 "exterior": [
            #                     [
            #                         2237,
            #                         1046
            #                     ],
            #                     [
            #                         2464,
            #                         1179
            #                     ]
            #                 ],
            #                 "interior": []
            #             }
            #         }
            #     ]
            # ]
        """
        filters = [{"field": "id", "operator": "in", "value": ids}]
        figures_infos = self.get_list_all_pages(
            "figures.list", {ApiField.DATASET_ID: dataset_id, ApiField.FILTER: filters}
        )

        if len(ids) != len(figures_infos):
            ids_downloaded = [info.id for info in figures_infos]
            raise RuntimeError(
                "Ids don't exist on server: {}".format(set(ids_downloaded) - set(ids))
            )

        id_to_item = {info.id: info for info in figures_infos}

        figures = []
        for input_id in ids:
            figures.append(id_to_item[input_id])

        return figures

    def _append_bulk(
        self,
        entity_id,
        figures_json,
        figures_keys,
        key_id_map: KeyIdMap,
        field_name=ApiField.ENTITY_ID,
    ):
        """"""
        if len(figures_json) == 0:
            return
        for (batch_keys, batch_jsons) in zip(
            batched(figures_keys, batch_size=100), batched(figures_json, batch_size=100)
        ):
            resp = self._api.post(
                "figures.bulk.add",
                {field_name: entity_id, ApiField.FIGURES: batch_jsons},
            )
            for key, resp_obj in zip(batch_keys, resp.json()):
                figure_id = resp_obj[ApiField.ID]
                key_id_map.add_figure(key, figure_id)



class DialogWindowError(Exception):
    def __init__(self, title, description):
        self.title = title
        self.description = description
        super().__init__(self.get_message())

    def get_message(self):
        return f"{self.title}: {self.description}"

    def __str__(self):
        return self.get_message()



# coding: utf-8
"""Tag can be attached to whole image and/or to individual :class:`Label<supervisely.annotation.label.Label>`"""

# docs
from __future__ import annotations
from typing import List, Optional, Dict, Union
from supervisely.annotation.tag_meta import TagMeta
from supervisely.annotation.tag_meta_collection import TagMetaCollection


from supervisely.annotation.tag_meta import TagValueType
from supervisely.collection.key_indexed_collection import KeyObject
from supervisely._utils import take_with_default


class TagJsonFields:
    """Json fields for :class:`Annotation<supervisely.annotation.tag.Tag>`"""

    TAG_NAME = "name"
    """"""
    VALUE = "value"
    """"""
    LABELER_LOGIN = "labelerLogin"
    """"""
    UPDATED_AT = "updatedAt"
    """"""
    CREATED_AT = "createdAt"
    """"""
    ID = "id"
    """"""
    # TAG_META_ID = 'tagId'
    # """"""


class Tag(KeyObject):
    """
    :class:`Tag<Tag>` can be attached to whole image and/or to individual :class:`Label<LabelBase>`. :class:`Tag<Tag>` object is immutable.

    :param meta: General information about Tag.
    :type meta: TagMeta
    :param value: Tag value. Depends on :class:`TagValueType<TagValueType>` of :class:`TagMeta<TagMeta>`.
    :type value: str or int or float or None
    :param sly_id: Tag ID in Supervisely server.
    :type sly_id: int, optional
    :param labeler_login: Login of user who created Tag.
    :type labeler_login: str, optional
    :param updated_at: Date and Time when Tag was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
    :type updated_at: str, optional
    :param created_at: Date and Time when Tag was created. Date Format is the same as in "updated_at" parameter.
    :type created_at: str, optional
    :raises: :class:`ValueError`, if meta is None or if Tag has incompatible value against it's meta value type
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        # Let's create 3 Tags with different values
        # First we need to initialize a TagMeta
        meta_dog = sly.TagMeta('dog', sly.TagValueType.NONE)

        # Now we can create a Tag using our TagMeta
        tag_dog = sly.Tag(meta_dog)

        # When you are creating a new Tag
        # Tag.value is automatically cross-checked against your TagMeta value type to make sure the value is valid.
        # If we now try to add a value to our newly created Tag, we receive "ValueError", because our TagMeta value type is "NONE"
        tag_dog = sly.Tag(meta_dog, value="Husky")
        # Output: ValueError: Tag dog can not have value Husky

        # Let's create another Tag with a string value type
        meta_cat = sly.TagMeta('cat', sly.TagValueType.ANY_STRING)
        tag_cat = sly.Tag(meta_cat, value="Fluffy")

        # Now let's create a Tag using TagMeta with "ONEOF_STRING" value type
        # In order to use "oneof_string value type", you must initialize a variable with possible values(see class TagMeta for more information)
        colors = ["brown", "white", "black", "red", "chocolate", "gold", "grey"]
        meta_coat_color = sly.TagMeta('coat color', sly.TagValueType.ONEOF_STRING, possible_values=colors)
        tag_coat_color = sly.Tag(meta_coat_color, value="white")

        # If given value is not in a list of possible Tags, ValueError will be raised
        tag_coat_color = sly.Tag(meta_coat_color, value="yellow")
        # Output: ValueError: Tag coat color can not have value yellow
    """

    def __init__(
        self,
        meta: TagMeta,
        value: Optional[Union[str, int, float]] = None,
        sly_id: Optional[int] = None,
        labeler_login: Optional[str] = None,
        updated_at: Optional[str] = None,
        created_at: Optional[str] = None,
    ):
        if meta is None:
            raise ValueError("TagMeta is None")
        self._meta = meta
        self._value = value
        if not self._meta.is_valid_value(value):
            raise ValueError("Tag {} can not have value {}".format(self.meta.name, value))
        self.labeler_login = labeler_login
        self.updated_at = updated_at
        self.created_at = created_at
        self.sly_id = sly_id

    @property
    def meta(self) -> TagMeta:
        """
        General information about Tag. When creating a new Tag, it's value is automatically cross-checked against :class:`TagValueType<supervisely.annotation.tag_meta.TagValueType>` to make sure that value is valid.

        :return: TagMeta object
        :rtype: :class:`TagMeta<supervisely.annotation.tag_meta.TagMeta>`
        :Usage example:

         .. code-block:: python

            meta_dog = sly.TagMeta('dog', sly.TagValueType.NONE)
            tag_dog = sly.Tag(meta_dog)

            # Our TagMeta has value type 'NONE', if we try to add value to our Tag, "ValueError" error will be raised
            tag_dog = sly.Tag(meta_dog, value="Husky")
            # Output: ValueError: Tag dog can not have value Husky
        """
        return self._meta

    @property
    def value(self) -> str or int or float:
        """
        Tag value. Return type depends on :class:`TagValueType<supervisely.annotation.tag_meta.TagValueType>`.

        :return: Tag value
        :rtype: :class:`str`, :class:`int` or :class:`float` or :class:`None`
        :Usage example:

         .. code-block:: python

            meta_dog = sly.TagMeta('dog', sly.TagValueType.ANY_STRING)
            tag_dog = sly.Tag(meta_dog, value="Husky")

            meta_age = sly.TagMeta('age', sly.TagValueType.ANY_NUMBER)
            tag_age = sly.Tag(meta_age, value=9)

            colors = ["Black", "White", "Golden", "Brown"]
            meta_color = sly.TagMeta('coat color', sly.TagValueType.ONEOF_STRING, possible_values=colors)
            tag_color = sly.Tag(meta_color, value="White")

            type(tag_dog.value)   # 'str'
            type(tag_age.value)   # 'int'
            type(tag_color.value) # 'str'
        """
        return self._value

    @property
    def name(self) -> str:
        """
        Name.

        :return: Name
        :rtype: :class:`str`
        :Usage example:

         .. code-block:: python

            meta_dog = sly.TagMeta('dog', sly.TagValueType.ANY_STRING)
            tag_dog = sly.Tag(meta_dog, value="Husky")

            print(tag_dog.name)
            # Output: "dog"
        """
        return self._meta.name

    def key(self):
        """ """
        return self._meta.key()

    def to_json(self) -> Dict:
        """
        Convert the Tag to a json dict. Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :return: Json format as a dict
        :rtype: :class:`dict`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            # Tag with all fields filled in
            meta_dog = sly.TagMeta('dog', sly.TagValueType.ANY_STRING)
            tag_dog = sly.Tag(meta=meta_dog, value="Husky", sly_id=38456, labeler_login="admin",
                              updated_at="2021-01-22T19:37:50.158Z", created_at="2021-01-22T18:00:00.000Z")

            tag_dog_json = tag_dog.to_json()
            print(tag_dog_json)
            # Notice that sly_id won't print
            # Output: {
            #  "name": "dog",
            #  "value": "Husky",
            #  "labelerLogin": "admin",
            #  "updatedAt": "2021-01-22T19:37:50.158Z",
            #  "createdAt": "2021-01-22T18:00:00.000Z"
            # }
        """
        res = {
            TagJsonFields.TAG_NAME: self.meta.name,
            # TagJsonFields.VALUE: self.value
        }
        if self.meta.value_type != TagValueType.NONE:
            res[TagJsonFields.VALUE] = self.value
        if self.labeler_login is not None:
            res[TagJsonFields.LABELER_LOGIN] = self.labeler_login
        if self.updated_at is not None:
            res[TagJsonFields.UPDATED_AT] = self.updated_at
        if self.created_at is not None:
            res[TagJsonFields.CREATED_AT] = self.created_at
        # if self.meta.value_type == TagValueType.NONE:
        #     return self.meta.name
        # else:
        #     return {
        #         TagJsonFields.TAG_NAME: self.meta.name,
        #         TagJsonFields.VALUE: self.value
        #     }
        return res

    @classmethod
    def from_json(cls, data: Dict, tag_meta_collection: TagMetaCollection) -> Tag:
        """
        Convert a json dict to Tag. Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :param data: Tag in json format as a dict.
        :type data: dict
        :param tag_meta_collection: TagMetaCollection object.
        :type tag_meta_collection: TagMetaCollection
        :return: Tag object
        :rtype: :class:`Tag<Tag>`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            meta_dog = sly.TagMeta('dog', sly.TagValueType.ANY_STRING)
            tag_metas = sly.TagMetaCollection([meta_dog])

            data = {
                "name": "dog",
                "value": "Husky",
                "labelerLogin": "admin",
                "updatedAt": "2021-01-22T19:37:50.158Z",
                "createdAt": "2021-01-22T18:00:00.000Z"
            }

            tag_dog = sly.Tag.from_json(data, tag_metas)
        """
        if type(data) is str:
            tag_name = data
            value = None
            labeler_login = None
            updated_at = None
            created_at = None
            sly_id = None
        else:
            tag_name = data[TagJsonFields.TAG_NAME]
            value = data.get(TagJsonFields.VALUE, None)
            labeler_login = data.get(TagJsonFields.LABELER_LOGIN, None)
            updated_at = data.get(TagJsonFields.UPDATED_AT, None)
            created_at = data.get(TagJsonFields.CREATED_AT, None)
            sly_id = data.get(TagJsonFields.ID, None)
        meta = tag_meta_collection.get(tag_name)
        return cls(
            meta=meta,
            value=value,
            sly_id=sly_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )

    def get_compact_str(self) -> str:
        """
        Displays information about Tag's name and value in string format.

        :return: Name and value of the given Tag
        :rtype: :class:`str`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            meta_dog = sly.TagMeta('dog', sly.TagValueType.ANY_STRING)
            tag_dog = sly.Tag(meta_dog, value="Husky")

            print(tag_dog.get_compact_str())
            # Output: 'dog:Husky'
        """
        if (self.meta.value_type != TagValueType.NONE) and (len(str(self.value)) > 0):
            return "{}:{}".format(self.name, self.value)
        return self.name

    def __eq__(self, other: Tag) -> bool:
        """
        Checks that 2 Tags are equal by comparing their meta and value.

        :param other: Tag object.
        :type other: Tag
        :return: True if comparable objects are equal, otherwise False
        :rtype: :class:`bool`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            # Let's create 2 identical Tags
            meta_lemon_1 = sly.TagMeta('Lemon', sly.TagValueType.NONE)
            tag_lemon_1 = sly.Tag(meta_lemon_1)

            meta_lemon_2 = sly.TagMeta('Lemon', sly.TagValueType.NONE)
            tag_lemon_2 = sly.Tag(meta_lemon_2)

            # and 1 different Tag to compare them
            meta_cucumber = sly.TagMeta('Cucumber', sly.TagValueType.ANY_STRING)
            tag_cucumber = sly.Tag(meta_cucumber, value="Fresh")

            # Compare identical Tags
            tag_lemon_1 == tag_lemon_2      # True

            # Compare unidentical Tags
            tag_lemon_1 == tag_cucumber     # False
        """
        return isinstance(other, Tag) and self.meta == other.meta and self.value == other.value

    def __ne__(self, other: Tag) -> bool:
        """
        Checks that 2 Tags are opposite.

        :param other: Tag object.
        :type other: Tag
        :return: True if comparable objects are not equal, otherwise False
        :rtype: :class:`bool`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            # Let's create 2 identical Tags
            meta_lemon_1 = sly.TagMeta('Lemon', sly.TagValueType.NONE)
            tag_lemon_1 = sly.Tag(meta_lemon_1)

            meta_lemon_2 = sly.TagMeta('Lemon', sly.TagValueType.NONE)
            tag_lemon_2 = sly.Tag(meta_lemon_2)

            # and 1 different Tag to compare them
            meta_cucumber = sly.TagMeta('Cucumber', sly.TagValueType.ANY_STRING)
            tag_cucumber = sly.Tag(meta_cucumber, value="Fresh")

            # Compare identical Tags
            tag_lemon_1 != tag_lemon_2      # False

            # Compare unidentical Tags
            tag_lemon_1 != tag_cucumber     # True
        """
        return not self == other

    def clone(
        self,
        meta: Optional[TagMeta] = None,
        value: Optional[Union[str, int, float]] = None,
        sly_id: Optional[int] = None,
        labeler_login: Optional[str] = None,
        updated_at: Optional[str] = None,
        created_at: Optional[str] = None,
    ) -> Tag:
        """
        Clone makes a copy of Tag with new fields, if fields are given, otherwise it will use original Tag fields.

        :param meta: General information about Tag.
        :type meta: TagMeta
        :type value: str or int or float or None
        :param sly_id: Tag ID in Supervisely server.
        :type sly_id: int, optional
        :param labeler_login: Login of user who created Tag.
        :type labeler_login: str, optional
        :param updated_at: Date and Time when Tag was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
        :type updated_at: str, optional
        :param created_at: Date and Time when Tag was created. Date Format is the same as in "updated_at" parameter.
        :type created_at: str, optional
        :return: New instance of Tag
        :rtype: :class:`Tag<Tag>`
        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            #Original Tag
            weather_conditions = ["Sunny", "Cloudy", "Snowy", "Foggy", "Rainy"]
            meta_weather = sly.TagMeta("weather", sly.TagValueType.ONEOF_STRING, possible_values=weather_conditions)

            tag_weather = sly.Tag(meta_weather, value="Sunny")

            # Let's create some more tags by cloning our original Tag
            # Remember that Tag class object is immutable, and we need to assign new instance of Tag to a new variable
            clone_weather_1 = tag_weather.clone(value="Snowy")

            clone_weather_2 = tag_weather.clone(value="Cloudy")

            clone_weather_3 = tag_weather.clone(value="Rainy")
        """
        return Tag(
            meta=take_with_default(meta, self.meta),
            value=take_with_default(value, self.value),
            sly_id=take_with_default(sly_id, self.sly_id),
            labeler_login=take_with_default(labeler_login, self.labeler_login),
            updated_at=take_with_default(updated_at, self.updated_at),
            created_at=take_with_default(created_at, self.created_at),
        )

    def __str__(self):
        return "{:<7s}{:<10}{:<7s} {:<13}{:<7s} {:<10}".format(
            "Name:",
            self._meta.name,
            "Value type:",
            self._meta.value_type,
            "Value:",
            str(self.value),
        )

    @classmethod
    def get_header_ptable(cls):
        """ """
        return ["Name", "Value type", "Value"]

    def get_row_ptable(self):
        """ """
        return [self._meta.name, self._meta.value_type, self.value]

    def _set_id(self, id: int):
        self.sly_id = id

    def _set_updated_at(self, updated_at: str):
        self.updated_at = updated_at

    def _set_created_at(self, created_at: str):
        self.created_at = created_at

    def _set_labeler_login(self, labeler_login: str):
        self.labeler_login = labeler_login



# coding: utf-8

# docs
from collections import namedtuple
import os
from typing import List, Optional, Tuple
from supervisely.api.api import Api


from supervisely.io.fs import file_exists, touch
from supervisely.io.json import dump_json_file, load_json_file
from supervisely.project.project_meta import ProjectMeta
from supervisely.task.progress import Progress
from supervisely._utils import batched
from supervisely.video_annotation.key_id_map import KeyIdMap

from supervisely.api.module_api import ApiField
from supervisely.collection.key_indexed_collection import KeyIndexedCollection
from supervisely.video import video as sly_video

from supervisely.project.project import Dataset, Project, OpenMode
from supervisely.project.project import read_single_project as read_project_wrapper
from supervisely.project.project_type import ProjectType
from supervisely.video_annotation.video_annotation import VideoAnnotation

VideoItemPaths = namedtuple("VideoItemPaths", ["video_path", "ann_path"])


class VideoDataset(Dataset):
    """
    This is a class for creating and using VideoDataset objects. Here is where your labeled and unlabeled images and other
    files live. There is no more levels: images or videos are directly attached to a dataset. Dataset is a unit of work.
    All images or videos are directly attached to a dataset. A dataset is some sort of data folder with stuff to annotate.
    """

    item_dir_name = "video"
    annotation_class = VideoAnnotation
    item_module = sly_video
    paths_tuple = VideoItemPaths

    @classmethod
    def _has_valid_ext(cls, path: str) -> bool:
        """
        Checks if file from given path is supported
        :param path: str
        :return: bool
        """
        return cls.item_module.has_valid_ext(path)

    def _get_empty_annotaion(self, item_name):
        """
        Create empty VideoAnnotation for given video
        :param item_name: str
        :return: VideoAnnotation class object
        """
        (
            img_size,
            frames_count,
        ) = VideoDataset.item_module.get_image_size_and_frames_count(item_name)
        return self.annotation_class(img_size, frames_count)

    def add_item_np(self, item_name, img, ann=None):
        raise RuntimeError("Deprecated method. Works only with images project")

    def _add_img_np(self, item_name, img):
        raise RuntimeError("Deprecated method. Works only with images project")

    @staticmethod
    def _validate_added_item_or_die(item_path):
        """
        Make sure we actually received a valid image file, clean it up and fail if not so.
        :param item_path: str
        """
        # Make sure we actually received a valid image file, clean it up and fail if not so.
        try:
            VideoDataset.item_module.validate_format(item_path)
        except Exception as e:
            os.remove(item_path)
            raise e

    def get_item_paths(self, item_name) -> VideoItemPaths:
        """
        :param item_name: str
        :return: namedtuple object containing paths to video and annotation from given path
        """
        return VideoDataset.paths_tuple(
            self.get_img_path(item_name), self.get_ann_path(item_name)
        )


class VideoProject(Project):
    """
    This is a class for creating and using VideoProject objects. You can think of a VideoProject as a superfolder with data and
    meta information.
    """

    dataset_class = VideoDataset

    class DatasetDict(KeyIndexedCollection):
        item_type = VideoDataset

    def __init__(self, directory, mode: OpenMode):
        """
        :param directory: path to the directory where the project will be saved or where it will be loaded from
        :param mode: OpenMode class object which determines in what mode to work with the project (generate exception error if not so)
        """
        self._key_id_map: KeyIdMap = None
        super().__init__(directory, mode)

    def _read(self):
        """
        Download project from given project directory. Checks item and annotation directoris existing and dataset not empty.
        Consistency checks. Every video must have an annotation, and the correspondence must be one to one.
        """
        super()._read()
        self._key_id_map = KeyIdMap()
        if os.path.exists(self._get_key_id_map_path()):
            self._key_id_map.load_json(self._get_key_id_map_path())

    def _create(self):
        """
        Creates a leaf directory and empty meta.json file. Generate exception error if project directory already exists and is not empty.
        """
        super()._create()
        self.set_key_id_map(KeyIdMap())

    def _add_item_file_to_dataset(
        self, ds, item_name, item_paths, _validate_item, _use_hardlink
    ):
        """
        Add given item file to dataset items directory and add annatation to dataset annotations dir corresponding to item.
        Generate exception error if item_name already exists in dataset or item name has unsupported extension
        :param ds: VideoDataset class object
        :param item_name: str
        :param item_paths: ItemPaths object
        :param _validate_item: bool
        :param _use_hardlink: bool
        """
        ds.add_item_file(
            item_name,
            item_paths.item_path,
            ann=item_paths.ann_path,
            _validate_item=_validate_item,
            _use_hardlink=_use_hardlink,
        )

    @property
    def key_id_map(self):
        return self._key_id_map

    def set_key_id_map(self, new_map: KeyIdMap):
        """
        Save given KeyIdMap object to project dir in json format.
        :param new_map: KeyIdMap class object
        """
        self._key_id_map = new_map
        self._key_id_map.dump_json(self._get_key_id_map_path())

    def _get_key_id_map_path(self):
        """
        :return: str (full path to key_id_map.json)
        """
        return os.path.join(self.directory, "key_id_map.json")

    @classmethod
    def read_single(cls, dir):
        """
        Read project from given ditectory. Generate exception error if given dir contains more than one subdirectory
        :param dir: str
        :return: VideoProject class object
        """
        return read_project_wrapper(dir, cls)


def download_video_project(
    api: Api,
    project_id: int,
    dest_dir: str,
    dataset_ids: List[int] = None,
    download_videos: bool = True,
    log_progress: bool = False,
) -> None:
    """
    Download project with given id in destination directory.

    :param api: Api class object.
    :type api: Api
    :param project_id: Project ID in Supervisely.
    :type project_id: int
    :param dest_dir: Directory to download video project.
    :type dest_dir: str
    :param dataset_ids: Datasets IDs in Supervisely to download.
    :type dataset_ids: List[int], optional
    :param download_videos: Download videos from Supervisely video project in dest_dir or not.
    :type download_videos: bool, optional
    :param log_progress: Logging progress of download video project or not.
    :type log_progress: bool, optional
    :return: None
    :rtype: :class:`NoneType`
    """
    LOG_BATCH_SIZE = 1

    key_id_map = KeyIdMap()

    project_fs = VideoProject(dest_dir, OpenMode.CREATE)

    meta = ProjectMeta.from_json(api.project.get_meta(project_id))
    project_fs.set_meta(meta)

    datasets_infos = []
    if dataset_ids is not None:
        for ds_id in dataset_ids:
            datasets_infos.append(api.dataset.get_info_by_id(ds_id))
    else:
        datasets_infos = api.dataset.get_list(project_id)

    for dataset in datasets_infos:
        dataset_fs = project_fs.create_dataset(dataset.name)
        videos = api.video.get_list(dataset.id)

        ds_progress = None
        if log_progress:
            ds_progress = Progress(
                "Downloading dataset: {!r}".format(dataset.name), total_cnt=len(videos)
            )
        for batch in batched(videos, batch_size=LOG_BATCH_SIZE):
            video_ids = [video_info.id for video_info in batch]
            video_names = [video_info.name for video_info in batch]

            ann_jsons = api.video.annotation.download_bulk(dataset.id, video_ids)

            for video_id, video_name, ann_json in zip(
                video_ids, video_names, ann_jsons
            ):
                if video_name != ann_json[ApiField.VIDEO_NAME]:
                    raise RuntimeError(
                        "Error in api.video.annotation.download_batch: broken order"
                    )

                video_file_path = dataset_fs.generate_item_path(video_name)
                if download_videos is True:
                    api.video.download_path(video_id, video_file_path)
                else:
                    touch(video_file_path)

                dataset_fs.add_item_file(
                    video_name,
                    video_file_path,
                    ann=VideoAnnotation.from_json(
                        ann_json, project_fs.meta, key_id_map
                    ),
                    _validate_item=False,
                )

            if log_progress:
                ds_progress.iters_done_report(len(batch))

    project_fs.set_key_id_map(key_id_map)


def upload_video_project(
    dir: str,
    api: Api,
    workspace_id: int,
    project_name: Optional[str] = None,
    log_progress: Optional[bool] = True,
) -> Tuple[int, str]:
    """
    Upload video project from given directory in Supervisely.

    :param dir: Directory with video project.
    :type dir: str
    :param api: Api class object.
    :type api: Api
    :param workspace_id: Workspace ID in Supervisely to upload video project.
    :type workspace_id: int
    :param project_name: Name of video project.
    :type project_name: str
    :param log_progress: Logging progress of download video project or not.
    :type log_progress: bool, optional
    :return: New video project ID in Supervisely and project name
    :rtype: :class:`Tuple[int, str]`
    """
    project_fs = VideoProject.read_single(dir)
    if project_name is None:
        project_name = project_fs.name

    if api.project.exists(workspace_id, project_name):
        project_name = api.project.get_free_name(workspace_id, project_name)

    project = api.project.create(workspace_id, project_name, ProjectType.VIDEOS)
    api.project.update_meta(project.id, project_fs.meta.to_json())

    for dataset_fs in project_fs.datasets:
        dataset = api.dataset.create(project.id, dataset_fs.name)

        names, item_paths, ann_paths = [], [], []
        for item_name in dataset_fs:
            img_path, ann_path = dataset_fs.get_item_paths(item_name)
            names.append(item_name)
            item_paths.append(img_path)
            ann_paths.append(ann_path)

        progress_cb = None
        if log_progress:
            ds_progress = Progress(
                "Uploading videos to dataset {!r}".format(dataset.name),
                total_cnt=len(item_paths),
            )
            progress_cb = ds_progress.iters_done_report

        item_infos = api.video.upload_paths(dataset.id, names, item_paths, progress_cb)
        item_ids = [item_info.id for item_info in item_infos]
        if log_progress:
            ds_progress = Progress(
                "Uploading annotations to dataset {!r}".format(dataset.name),
                total_cnt=len(item_paths),
            )
            progress_cb = ds_progress.iters_done_report

        api.video.annotation.upload_paths(
            item_ids, ann_paths, project_fs.meta, progress_cb
        )

    return project.id, project.name



# coding: utf-8


# docs
from __future__ import annotations
from typing import List, Dict, Optional, Iterator, Any
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.annotation.tag_meta_collection import TagMetaCollection

from supervisely.annotation.tag_collection import TagCollection
from supervisely.video_annotation.video_tag import VideoTag


class VideoTagCollection(TagCollection):
    """
    Collection with :class:`VideoTag<supervisely.video_annotation.video_tag.VideoTag>` instances. :class:`VideoTagCollection<VideoTagCollection>` object is immutable.

    :Usage example:

     .. code-block:: python

        import supervisely as sly
        from supervisely.video_annotation.video_tag import VideoTag
        from supervisely.video_annotation.video_tag_collection import VideoTagCollection

        # Create two VideoTags for collection
        meta_car = sly.TagMeta('car_tag', sly.TagValueType.ANY_STRING)
        car_tag = VideoTag(meta_car, value='acura')
        meta_bus = sly.TagMeta('bus_tag', sly.TagValueType.ANY_STRING)
        bus_tag = VideoTag(meta_bus, value='volvo')

        # Create VideoTagCollection
        tags = VideoTagCollection([car_tag, bus_tag])
        tags_json = tags.to_json()
        print(tags_json)
        # Output: [
        #     {
        #         "name": "car_tag",
        #         "value": "acura",
        #         "key": "378408fcb6854305a38fed7c996f4901"
        #     },
        #     {
        #         "name": "bus_tag",
        #         "value": "volvo",
        #         "key": "0c63174878204faea67c4025adec1e2a"
        #     }
        # ]

        # Add item to VideoTagCollection
        meta_truck = sly.TagMeta('truck_tag', sly.TagValueType.NONE)
        truck_tag = VideoTag(meta_truck)
        # Remember that VideoTagCollection is immutable, and we need to assign new instance of VideoTagCollection to a new variable
        new_tags = tags.add(truck_tag)
        new_tags_json = new_tags.to_json()
        print(new_tags_json)
        # Output: [
        #     {
        #         "name": "car_tag",
        #         "value": "acura",
        #         "key": "3d33f1685dab44da9b55d67bab3937e9"
        #     },
        #     {
        #         "name": "bus_tag",
        #         "value": "volvo",
        #         "key": "7c52c168a22b47a5b39fbcdef94b0140"
        #     },
        #     {
        #         "name": "truck_tag",
        #         "key": "7188242d230b4d2783c588cc2eca5ff8"
        #     }
        # ]

        # You can also add multiple items to collection
        meta_truck = sly.TagMeta('truck_tag', sly.TagValueType.NONE)
        truck_tag = VideoTag(meta_truck)
        meta_train = sly.TagMeta('train_tag', sly.TagValueType.ANY_NUMBER)
        train_tag = VideoTag(meta_train, value=777)
        new_tags = tags.add_items([truck_tag, train_tag])
        new_tags_json = new_tags.to_json()
        print(new_tags_json)
        # Output: [
        #     {
        #         "name": "car_tag",
        #         "value": "acura",
        #         "key": "03b052c464d84d5db6451b86f3bdef79"
        #     },
        #     {
        #         "name": "bus_tag",
        #         "value": "volvo",
        #         "key": "0f7dd8fb6f8c41da9e68e64d3186df15"
        #     },
        #     {
        #         "name": "truck_tag",
        #         "key": "fc01af7c70154771b7253b6a94484179"
        #     },
        #     {
        #         "name": "train_tag",
        #         "value": 777,
        #         "key": "6ce5118181074a52b74dee9335fa292d"
        #     }
        # ]

        # Intersection, finds intersection of given list of VideoTag instances with collection items
        intersect_tags = tags.intersection([bus_tag])
        intersect_tags_json = intersect_tags.to_json()
        print(intersect_tags_json)
        # Output: [
        #     {
        #         "name": "bus_tag",
        #         "value": "volvo",
        #         "key": "13d77d8c848e4a3ebeb710bf5f3f38a6"
        #     }
        # ]

        # Difference, finds difference between collection and given list of VideoTag
        diff_tags = tags.difference([bus_tag])
        diff_tags_json = diff_tags.to_json()
        print(diff_tags_json)
        # Output: [
        #     {
        #         "name": "car_tag",
        #         "value": "acura",
        #         "key": "341b9fc077e142c0956d5cf985d705c1"
        #     }
        # ]

        # Merge, merges collection and given list of VideoTagCollection
        meta_truck = sly.TagMeta('truck_tag', sly.TagValueType.NONE)
        truck_tag = VideoTag(meta_truck)
        meta_train = sly.TagMeta('train_tag', sly.TagValueType.ANY_NUMBER)
        train_tag = VideoTag(meta_train, value=777)
        over_tags = VideoTagCollection([truck_tag, train_tag])
        # Merge
        merge_tags = tags.merge(over_tags)
        merge_tags_json = merge_tags.to_json()
        print(merge_tags_json)
        # Output: [
        #     {
        #         "name": "car_tag",
        #         "value": "acura",
        #         "key": "3bfde60091a24dd493485f2b80364736"
        #     },
        #     {
        #         "name": "bus_tag",
        #         "value": "volvo",
        #         "key": "dac5b07954064c78acadfa71161f6998"
        #     },
        #     {
        #         "name": "truck_tag",
        #         "key": "a46aac4499574afc81d41d2bb061a3c6"
        #     },
        #     {
        #         "name": "train_tag",
        #         "value": 777,
        #         "key": "bf5b1667383449478694fb6349d7b16c"
        #     }
        # ]
    """

    item_type = VideoTag

    def to_json(self, key_id_map: Optional[KeyIdMap] = None) -> List[Dict]:
        """
        Convert the VideoTagCollection to a list of json dicts. Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :return: List of dicts in json format
        :rtype: :class:`List[dict]`
        :Usage example:

         .. code-block:: python

            import supervisely as sly
            from supervisely.video_annotation.video_tag import VideoTag
            from supervisely.video_annotation.video_tag_collection import VideoTagCollection

            meta_car = sly.TagMeta('car_tag', sly.TagValueType.ANY_STRING)
            car_tag = VideoTag(meta_car, value='acura')
            meta_bus = sly.TagMeta('bus_tag', sly.TagValueType.ANY_STRING)
            bus_tag = VideoTag(meta_bus, value='volvo')
            tags = VideoTagCollection([car_tag, bus_tag])
            tags_json = tags.to_json()
            print(tags_json)
            # Output: [
            #     {
            #         "name": "car_tag",
            #         "value": "acura",
            #         "key": "378408fcb6854305a38fed7c996f4901"
            #     },
            #     {
            #         "name": "bus_tag",
            #         "value": "volvo",
            #         "key": "0c63174878204faea67c4025adec1e2a"
            #     }
            # ]
        """
        return [tag.to_json(key_id_map) for tag in self]

    @classmethod
    def from_json(
        cls,
        data: List[Dict],
        tag_meta_collection: TagMetaCollection,
        key_id_map: Optional[KeyIdMap] = None,
    ) -> VideoTagCollection:
        """
        Convert a list of json dicts to VideoTagCollection. Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :param data: List with dicts in json format.
        :type data: List[dict]
        :param project_meta: Input TagMetaCollection object.
        :type project_meta: TagMetaCollection
        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :return: VideoTagCollection object
        :rtype: :class:`VideoTagCollection`

        :Usage example:

         .. code-block:: python

            import supervisely as sly
            from supervisely.video_annotation.video_tag_collection import VideoTagCollection

            tags_json = [
                {
                    "name": "car_tag",
                    "value": "acura",
                },
                {
                    "name": "bus_tag",
                    "value": "volvo",
                }
            ]
            meta_car = sly.TagMeta('car_tag', sly.TagValueType.ANY_STRING)
            meta_bus = sly.TagMeta('bus_tag', sly.TagValueType.ANY_STRING)
            meta_collection = sly.TagMetaCollection([meta_car, meta_bus])

            tags = VideoTagCollection.from_json(tags_json, meta_collection)
        """
        tags = [
            cls.item_type.from_json(tag_json, tag_meta_collection, key_id_map) for tag_json in data
        ]
        return cls(tags)

    def __iter__(self) -> Iterator[VideoTag]:
        return next(self)

    @classmethod
    def from_api_response(cls, data, tag_meta_collection, id_to_tagmeta=None) -> VideoTagCollection:
        return super().from_api_response(data, tag_meta_collection, id_to_tagmeta)

    def get_by_name(self, tag_name: str, default: Optional[Any] = None) -> List[VideoTag]:
        # super().get_by_name(tag_name, default)
        res = []
        for tag in self:
            if tag.name == tag_name:
                res.append(tag)
        return res

    def get_single_by_name(self, tag_name: str, default: Optional[Any] = None) -> VideoTag:
        # super().get_by_name(tag_name, default)
        res = []
        for tag in self:
            if tag.name == tag_name:
                res.append(tag)
        if len(res) == 0:
            return None
        if len(res) > 1:
            raise ValueError(
                "There are more than one tag {tag_name} in VideoTagCollection. Use method get_by_name instead"
            )
        return res[0]



# coding: utf-8
from __future__ import annotations

import random
import colorsys
from typing import List


def _validate_color(color):
    """
    Checks input color for compliance with the required format
    :param: color: color (RGB tuple of integers)
    """
    if not isinstance(color, (list, tuple)):
        raise ValueError('Color has to be list, or tuple')
    if len(color) != 3:
        raise ValueError('Color have to contain exactly 3 values: [R, G, B]')
    for channel in color:
        validate_channel_value(channel)


def random_rgb() -> List[int, int, int]:
    """
    Generate RGB color with fixed saturation and lightness.

    :return: RGB integer values
    :rtype: :class:`List[int, int, int]`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        color = sly.color.random_rgb()
        print(color)
        # Output: [138, 15, 123]
    """
    hsl_color = (random.random(), 0.3, 0.8)
    rgb_color = colorsys.hls_to_rgb(*hsl_color)
    return [round(c * 255) for c in rgb_color]


def _normalize_color(color):
    """
    Divide all RGB values by 255.
    :param color: color (RGB tuple of integers)
    """
    return [c / 255. for c in color]


def _color_distance(first_color: list, second_color: list) -> float:
    """
    Calculate distance in HLS color space between Hue components of 2 colors
    :param first_color: first color (RGB tuple of integers)
    :param second_color: second color (RGB tuple of integers)
    :return: Euclidean distance between 'first_color' and 'second_color'
    """
    first_color_hls = colorsys.rgb_to_hls(*_normalize_color(first_color))
    second_color_hls = colorsys.rgb_to_hls(*_normalize_color(second_color))
    hue_distance = min(abs(first_color_hls[0] - second_color_hls[0]),
                       1 - abs(first_color_hls[0] - second_color_hls[0]))
    return hue_distance


def generate_rgb(exist_colors: List[List[int, int, int]]) -> List[int, int, int]:
    """
    Generate new color which oppositely by exist colors.

    :param exist_colors: List of existing colors in RGB format.
    :type exist_colors: list
    :return: RGB integer values
    :rtype: :class:`List[int, int, int]`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        exist_colors = [[0, 0, 0], [128, 64, 255]]
        color = sly.color.generate_rgb(exist_colors)
        print(color)
        # Output: [15, 138, 39]
    """
    largest_min_distance = 0
    best_color = random_rgb()
    if len(exist_colors) > 0:
        for _ in range(100):
            color = random_rgb()
            current_min_distance = min(_color_distance(color, c) for c in exist_colors)
            if current_min_distance > largest_min_distance:
                largest_min_distance = current_min_distance
                best_color = color
    _validate_color(best_color)
    return best_color


def rgb2hex(color: List[int, int, int]) -> str:
    """
    Convert integer color format to HEX string.

    :param color: List of existing colors in RGB format.
    :type color: List[int, int, int]
    :return: HEX RGB string
    :rtype: :class:`str`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        hex_color = sly.color.rgb2hex([128, 64, 255])
        print(hex_color)
        # Output: #8040FF
    """
    _validate_color(color)
    return '#' + ''.join('{:02X}'.format(component) for component in color)


def _hex2color(hex_value: str) -> list:
    """
        Convert HEX RGB string to integer RGB format
        :param hex_value: HEX RGBA string. Example: "#FF02А4
        :return: RGB integer values. Example: [80, 255, 0]
    """
    assert hex_value.startswith('#')
    return [int(hex_value[i:(i + 2)], 16) for i in range(1, len(hex_value), 2)]


def hex2rgb(hex_value: str) -> List[int, int, int]:
    """
    Convert HEX RGB string to integer RGB format.

    :param hex_value: HEX RGB string.
    :type hex_value: str
    :return: RGB integer values
    :rtype: :class:`List[int, int, int]`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        hex_color = '#8040FF'
        color = sly.color.hex2rgb(hex_color)
        print(color)
        # Output: [128, 64, 255]
    """
    assert len(hex_value) == 7, "Supported only HEX RGB string format!"
    color = _hex2color(hex_value)
    _validate_color(color)
    return color


def _hex2rgba(hex_value: str) -> list:
    """
    Convert HEX RGBA string to integer RGBA format
    :param hex_value: HEX RGBA string. Example: "#FF02А4CC
    :return: RGBA integer values. Example: [80, 255, 0, 128]
    """
    assert len(hex_value) == 9, "Supported only HEX RGBA string format!"
    return _hex2color(hex_value)


def validate_channel_value(value: int) -> None:
    """
    Generates ValueError if value not between 0 and 255.

    :param value: Input channel value.
    :type value: int
    :raises: :class:`ValueError` if value not between 0 and 255.
    :return: None
    :rtype: :class:`NoneType`
    """
    if 0 <= value <= 255:
        pass
    else:
        raise ValueError('Color channel has to be in range [0; 255]')


from supervisely.app.singleton import Singleton


class JinjaWidgets(metaclass=Singleton):
    def __init__(self, auto_widget_id=True):
        self.auto_widget_id = auto_widget_id
        self.context = {}
        self.context["__no_html_mode__"] = auto_widget_id



# coding: utf-8

import os
from typing import NamedTuple, List
from supervisely.api.module_api import ApiField
from supervisely.api.task_api import TaskApi
from supervisely._utils import take_with_default

# from supervisely.app.constants import DATA, STATE, CONTEXT, TEMPLATE
STATE = "state"
DATA = "data"
TEMPLATE = "template"

from supervisely.io.fs import ensure_base_path
from supervisely.task.progress import Progress
from supervisely._utils import sizeof_fmt
from supervisely import logger


class AppInfo(NamedTuple):
    """AppInfo"""

    id: int
    created_by_id: int
    module_id: int
    disabled: bool
    user_login: str
    config: dict
    name: str
    slug: str
    is_shared: bool
    tasks: int
    repo: str
    team_id: int


class AppApi(TaskApi):
    """AppApi"""

    @staticmethod
    def info_sequence():
        """info_sequence"""
        return [
            ApiField.ID,
            ApiField.CREATED_BY_ID,
            ApiField.MODULE_ID,
            ApiField.DISABLED,
            ApiField.USER_LOGIN,
            ApiField.CONFIG,
            ApiField.NAME,
            ApiField.SLUG,
            ApiField.IS_SHARED,
            ApiField.TASKS,
            ApiField.REPO,
            ApiField.TEAM_ID,
        ]

    @staticmethod
    def info_tuple_name():
        """info_tuple_name"""
        return "AppInfo"

    def _convert_json_info(self, info: dict, skip_missing=True):
        """_convert_json_info"""
        res = super(TaskApi, self)._convert_json_info(info, skip_missing=skip_missing)
        return AppInfo(**res._asdict())

    def get_info_by_id(self, id: int) -> AppInfo:
        """
        :param id: int
        :return: application info by numeric id
        """
        return self._get_info_by_id(id, "apps.info")

    def get_list(
        self,
        team_id,
        filter=None,
        context=None,
        repository_key=None,
        show_disabled=False,
        integrated_into=None,
        session_tags=None,
        only_running=False,
        with_shared=True,
    ) -> List[AppInfo]:
        """get_list"""

        return self.get_list_all_pages(
            method="apps.list",
            data={
                "teamId": team_id,
                "filter": take_with_default(
                    filter, []
                ),  # for example [{"field": "id", "operator": "=", "value": None}]
                "context": take_with_default(
                    context, []
                ),  # for example ["images_project"]
                "repositoryKey": repository_key,
                "integratedInto": take_with_default(
                    integrated_into, []
                ),  # for example ["image_annotation_tool"]
                "sessionTags": take_with_default(
                    session_tags, []
                ),  # for example ["string"]
                "onlyRunning": only_running,
                "showDisabled": show_disabled,
                "withShared": with_shared,
            },
        )

    def run_dtl(self, workspace_id, dtl_graph, agent_id=None):
        """run_dtl"""
        raise RuntimeError("Method is unavailable")

    def _run_plugin_task(
        self,
        task_type,
        agent_id,
        plugin_id,
        version,
        config,
        input_projects,
        input_models,
        result_name,
    ):
        """_run_plugin_task"""
        raise RuntimeError("Method is unavailable")

    def run_train(
        self,
        agent_id,
        input_project_id,
        input_model_id,
        result_nn_name,
        train_config=None,
    ):
        """run_train"""
        raise RuntimeError("Method is unavailable")

    def run_inference(
        self,
        agent_id,
        input_project_id,
        input_model_id,
        result_project_name,
        inference_config=None,
    ):
        """run_inference"""
        raise RuntimeError("Method is unavailable")

    def get_training_metrics(self, task_id):
        """get_training_metrics"""
        raise RuntimeError("Method is unavailable")

    def deploy_model(self, agent_id, model_id):
        """deploy_model"""
        raise RuntimeError("Method is unavailable")

    def get_import_files_list(self, id):
        """get_import_files_list"""
        raise RuntimeError("Method is unavailable")

    def download_import_file(self, id, file_path, save_path):
        """download_import_file"""
        raise RuntimeError("Method is unavailable")

    def create_task_detached(self, workspace_id, task_type: str = None):
        """create_task_detached"""
        raise RuntimeError("Method is unavailable")

    def upload_files(self, task_id, abs_paths, names, progress_cb=None):
        """upload_files"""
        raise RuntimeError("Method is unavailable")

    def initialize(self, task_id, template, data=None, state=None):
        """initialize"""
        d = take_with_default(data, {})
        if "notifyDialog" not in d:
            d["notifyDialog"] = None
        if "scrollIntoView" not in d:
            d["scrollIntoView"] = None

        s = take_with_default(state, {})
        fields = [
            {"field": TEMPLATE, "payload": template},
            {"field": DATA, "payload": d},
            {"field": STATE, "payload": s},
        ]
        resp = self._api.task.set_fields(task_id, fields)
        return resp

    def get_url(self, task_id):
        """get_url"""
        return f"/apps/sessions/{task_id}"

    def download_git_file(self, app_id, version, file_path, save_path):
        """download_git_file"""
        raise NotImplementedError()

    def download_git_archive(
        self,
        ecosystem_item_id,
        app_id,
        version,
        save_path,
        log_progress=True,
        ext_logger=None,
    ):
        """download_git_archive"""
        payload = {
            ApiField.ECOSYSTEM_ITEM_ID: ecosystem_item_id,
            ApiField.VERSION: version,
            "isArchive": True,
        }
        if app_id is not None:
            payload[ApiField.APP_ID] = app_id

        response = self._api.post("ecosystem.file.download", payload, stream=True)
        if log_progress:
            if ext_logger is None:
                ext_logger = logger

            length = -1
            # Content-Length
            if "Content-Length" in response.headers:
                length = int(response.headers["Content-Length"])
            progress = Progress(
                "Downloading: ", length, ext_logger=ext_logger, is_size=True
            )

        mb1 = 1024 * 1024
        ensure_base_path(save_path)
        with open(save_path, "wb") as fd:
            log_size = 0
            for chunk in response.iter_content(chunk_size=mb1):
                fd.write(chunk)
                log_size += len(chunk)
                if log_progress and log_size > mb1:
                    progress.iters_done_report(log_size)
                    log_size = 0

    def get_info(self, module_id, version=None):
        """get_info"""
        data = {ApiField.ID: module_id}
        if version is not None:
            data[ApiField.VERSION] = version
        response = self._api.post("ecosystem.info", data)
        return response.json()

    def get_ecosystem_module_id(self, slug: str):
        modules = self.get_list_all_pages(
            method="ecosystem.list",
            data={"filter": [{"field": "slug", "operator": "=", "value": slug}]},
            convert_json_info_cb=lambda x: x,
        )
        if len(modules) == 0:
            raise KeyError(f"Module {slug} not found in ecosystem")
        if len(modules) > 1:
            raise KeyError(
                f"Ecosystem is broken: there are {len(modules)} modules with the same slug: {slug}. Please, contact tech support"
            )
        return modules[0]["id"]

    def get_sessions(
        self,
        team_id,
        module_id,
        # only_running=False,
        show_disabled=False,
        session_name=None,
    ):
        infos_json = self.get_list_all_pages(
            method="apps.list",
            data={
                "teamId": team_id,
                "filter": [{"field": "moduleId", "operator": "=", "value": module_id}],
                # "onlyRunning": only_running,
                "showDisabled": show_disabled,
            },
            convert_json_info_cb=lambda x: x,
            # validate_total=False,
        )
        if len(infos_json) == 0:
            raise KeyError(f"App [module_id = {module_id}] not found in team {team_id}")
        if len(infos_json) > 1:
            raise KeyError(
                f"Apps list in team is broken: app [module_id = {module_id}] added to team {team_id} multiple times"
            )
        dev_tasks = []
        sessions = infos_json[0]["tasks"]
        for session in sessions:
            if session["status"] in ["queued", "consumed", "started", "deployed"]:
                if session_name is not None:
                    if session["meta"]["name"] == session_name:
                        dev_tasks.append(session)
                else:
                    dev_tasks.append(session)
        return dev_tasks



# coding: utf-8

# fields names
BITMAP = 'bitmap'
COLS = 'cols'
DATA = 'data'
EXTERIOR = 'exterior'
FACES = 'faces'
INTERIOR = 'interior'
MULTICHANNEL_BITMAP = 'multichannelBitmap'
ORIGIN = 'origin'
POINTS = 'points'
ROWS = 'rows'
TYPE = 'type'
GEOMETRY_SHAPE = 'shape'
GEOMETRY_TYPE = 'geometryType'
ANY_SHAPE = 'any'

NODES = 'nodes'
EDGES = 'edges'
ENABLED = 'enabled'
LABEL = 'label'
COLOR = 'color'
TEMPLATE = 'template'
LOCATION = 'location'

X = "x"
Y = "y"
Z = "z"
POSITION = "position"
ROTATION = "rotation"
DIMENTIONS = "dimensions"


LABELER_LOGIN = 'labelerLogin'
UPDATED_AT = 'updatedAt'
CREATED_AT = 'createdAt'
ID = 'id'
CLASS_ID = 'classId'

INDICES = 'indices'
